/*
 * ungif.c -- Enfle yet another GIF handler, this uses libungif
 * (C)Copyright 1998, 99 by Hiroshi Takekawa
 * This file is part of Enfle.
 *
 * Last Modified: Sun Oct 17 02:27:03 1999.
 * $Id$
 *
 * NOTE: This file does not include LZW code.
 * NOTE: This plugin is incomplete, but works.
 *
 *             The Graphics Interchange Format(c) is
 *       the Copyright property of CompuServe Incorporated.
 * GIF(sm) is a Service Mark property of CompuServe Incorporated.
 *
 * Enfle is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Enfle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <gif_lib.h>
#include "enfle.h"
#include "ungif.h"
#include "plugin.h"

int
#ifdef PIC
get_plugininfo(PluginInfo *p)
#else
loader_ungif_get_plugininfo(PluginInfo *p)
#endif
{
  p->version = 1;
  p->type = _Loader;
  p->pluginname = "UNGIF Format Loader Plugin version 0.1";
  p->pluginshortname = FORMAT_NAME;
  p->author = "Hiroshi Takekawa";
  p->dlhandle = NULL; /* set by plugin_load */
  p->functions.loader.load_image = ungif_load_image;

  return 1;
}

static int
ungif_decode_image(Info *info, Image *p)
{
  int i, j, size, sheight, row, col, extcode;
  static int ioffset[] = { 0, 4, 2, 1 };
  static int ijumps[] = { 8, 8, 4, 2 };
  GifRecordType RecordType;
  GifByteType *Extension;
  GifRowType *ScreenBuffer;
  GifFileType *GifFile;
  ColorMapObject *ColorMap;

  if ((GifFile = DGifOpenFileName(info->filename)) == NULL) {
    /* PrintGifError(); */
    return PROC_NOT;
  }

  sheight = GifFile->SHeight;
  info->format = FORMAT_NAME;

  if ((ScreenBuffer = (GifRowType *)
       malloc(sheight * sizeof(GifRowType *))) == NULL) {
    if (DGifCloseFile(GifFile) == GIF_ERROR)
      PrintGifError();
    return PROC_ERROR;
  }    

  size = GifFile->SWidth * sizeof(GifPixelType);
  if ((ScreenBuffer[0] = (GifRowType)malloc(size)) == NULL) {
    if (DGifCloseFile(GifFile) == GIF_ERROR)
      PrintGifError();
    return PROC_ERROR;
  }

  for (i = 0; i < GifFile->SWidth; i++)
    ScreenBuffer[0][i] = GifFile->SBackGroundColor;
  for (i = 1; i < sheight; i++) {
    if ((ScreenBuffer[i] = (GifRowType)malloc(size)) == NULL) {
      if (DGifCloseFile(GifFile) == GIF_ERROR)
	PrintGifError();
      for (i--; i >= 0; i--)
	free(ScreenBuffer[i]);
      free(ScreenBuffer);
      return PROC_ERROR;
    }
    memcpy(ScreenBuffer[i], ScreenBuffer[0], size);
  }

  do {
#ifdef DEBUG
    fprintf(stderr, "GetRecordType ");
#endif
    if (DGifGetRecordType(GifFile, &RecordType) == GIF_ERROR) {
      PrintGifError();
#if 0
      DGifCloseFile(GifFile);
      for (i = sheight - 1; i >= 0; i--)
	free(ScreenBuffer[i]);
      free(ScreenBuffer);
      return PROC_ERROR;
#endif
    }
#ifdef DEBUG
    fprintf(stderr, "done.\n");
#endif
    switch (RecordType) {
    case IMAGE_DESC_RECORD_TYPE:
      if (DGifGetImageDesc(GifFile) == GIF_ERROR) {
	PrintGifError();
	DGifCloseFile(GifFile);
	for (i = sheight - 1; i >= 0; i--)
	  free(ScreenBuffer[i]);
	free(ScreenBuffer);
	return PROC_ERROR;
      }
      p->top = row = GifFile->Image.Top;
      p->left = col = GifFile->Image.Left;
      p->width = GifFile->Image.Width;
      p->height = GifFile->Image.Height;
#ifdef DEBUG
      fprintf(stderr, "IMAGE_DESC_RECORD_TYPE: (%d, %d) (%d, %d)\n",
	      p->left, p->top, p->width, p->height);
#endif
      if (GifFile->Image.Interlace) {
	for (i = 0; i < 4; i++)
	  for (j = row + ioffset[i]; j < row + p->height; j += ijumps[i])
	    if (DGifGetLine(GifFile, &ScreenBuffer[j][col], p->width) == GIF_ERROR) {
	      PrintGifError();
	      DGifCloseFile(GifFile);
	      for (i = sheight - 1; i >= 0; i--)
		free(ScreenBuffer[i]);
	      free(ScreenBuffer);
	      return PROC_ERROR;
	    }
      } else {
	for (i = 0; i < p->height; i++)
	  if (DGifGetLine(GifFile, &ScreenBuffer[row++][col], p->width) == GIF_ERROR) {
	    PrintGifError();
	    DGifCloseFile(GifFile);
	    for (i = sheight - 1; i >= 0; i--)
	      free(ScreenBuffer[i]);
	    free(ScreenBuffer);
	    return PROC_ERROR;
	  }
      }
      break;
    case EXTENSION_RECORD_TYPE:
      if (DGifGetExtension(GifFile, &extcode, &Extension) == GIF_ERROR) {
	PrintGifError();
	DGifCloseFile(GifFile);
	for (i = sheight - 1; i >= 0; i--)
	  free(ScreenBuffer[i]);
	free(ScreenBuffer);
	return PROC_ERROR;
      }
      while (Extension != NULL) {
	if (DGifGetExtensionNext(GifFile, &Extension) == GIF_ERROR) {
	  PrintGifError();
	  DGifCloseFile(GifFile);
	  for (i = sheight - 1; i >= 0; i--)
	    free(ScreenBuffer[i]);
	  free(ScreenBuffer);
	  return PROC_ERROR;
	}
      }
      break;
    default:
      break;
    }
  } while (RecordType != TERMINATE_RECORD_TYPE);

  ColorMap = GifFile->Image.ColorMap ? GifFile->Image.ColorMap : GifFile->SColorMap;
  p->ncolors = ColorMap->ColorCount;

  for (i = 0; i < p->ncolors; i++) {
    p->colormap[i][0] = ColorMap->Colors[i].Red;
    p->colormap[i][1] = ColorMap->Colors[i].Green;
    p->colormap[i][2] = ColorMap->Colors[i].Blue;
  }

  if (DGifCloseFile(GifFile) == GIF_ERROR) {
    PrintGifError();
    return PROC_ERROR;
  }

  p->type = _INDEX;
  p->bytes_per_line = p->width;
  p->image_size = p->bytes_per_line * p->height;
  if ((p->image = malloc(p->image_size)) == NULL) {
    for (i = sheight - 1; i >= 0; i--)
      free(ScreenBuffer[i]);
    free(ScreenBuffer);
    return PROC_ERROR;
  }

  for (i = 0; i < p->height; i++)
    for (j = 0; j < p->width; j++)
      p->image[i * p->width + j] = ScreenBuffer[p->top + i][p->left + j];

  for (i = sheight - 1; i >= 0; i--)
    free(ScreenBuffer[i]);
  free(ScreenBuffer);

  return PROC_OK;
}

int
ungif_load_image(Info *info, Image *p)
{
  int c;

  /* decode */
  if ((c = ungif_decode_image(info, p)) != PROC_OK)
    return c;

  /* set handler */
  info->handler.destroy  = NULL;
  info->handler.alarmset = NULL;
  info->handler.alarm    = NULL;
  info->handler.alarmoff = NULL;

  info->npics = 1;

  return PROC_OK;
}
